@@ -2,7 +2,7 @@
 
   range.c -
 
-  $Author: hsbt $
+  $Author$
   created at: Thu Aug 19 17:46:47 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -11,6 +11,7 @@
 
 #include "internal.h"
 #include "id.h"
+#include <assert.h>
 
 #ifdef HAVE_FLOAT_H
 #include <float.h>
@@ -18,15 +19,12 @@
 #include <math.h>
 
 VALUE rb_cRange;
-static ID id_beg, id_end, id_excl, id_integer_p, id_div;
+static ID id_beg, id_end, id_excl;
 #define id_cmp idCmp
 #define id_succ idSucc
 
 static VALUE r_cover_p(VALUE, VALUE, VALUE, VALUE);
 
-#define RANGE_BEG(r) (RSTRUCT(r)->as.ary[0])
-#define RANGE_END(r) (RSTRUCT(r)->as.ary[1])
-#define RANGE_EXCL(r) (RSTRUCT(r)->as.ary[2])
 #define RANGE_SET_BEG(r, v) (RSTRUCT_SET(r, 0, v))
 #define RANGE_SET_END(r, v) (RSTRUCT_SET(r, 1, v))
 #define RANGE_SET_EXCL(r, v) (RSTRUCT_SET(r, 2, v))
@@ -34,33 +32,15 @@ static VALUE r_cover_p(VALUE, VALUE, VALUE, VALUE);
 
 #define EXCL(r) RTEST(RANGE_EXCL(r))
 
-static VALUE
-range_failed(void)
-{
-    rb_raise(rb_eArgError, "bad value for range");
-    return Qnil;		/* dummy */
-}
-
-static VALUE
-range_check(VALUE *args)
-{
-    return rb_funcall(args[0], id_cmp, 1, args[1]);
-}
-
 static void
 range_init(VALUE range, VALUE beg, VALUE end, VALUE exclude_end)
 {
-    VALUE args[2];
-
-    args[0] = beg;
-    args[1] = end;
-
-    if (!FIXNUM_P(beg) || !FIXNUM_P(end)) {
+    if ((!FIXNUM_P(beg) || !FIXNUM_P(end)) && !NIL_P(beg) && !NIL_P(end)) {
 	VALUE v;
 
-	v = rb_rescue(range_check, (VALUE)args, range_failed, 0);
+	v = rb_funcall(beg, id_cmp, 1, end);
 	if (NIL_P(v))
-	    range_failed();
+	    rb_raise(rb_eArgError, "bad value for range");
     }
 
     RANGE_SET_EXCL(range, exclude_end);
@@ -80,6 +60,7 @@ rb_range_new(VALUE beg, VALUE end, int exclude_end)
 static void
 range_modify(VALUE range)
 {
+    rb_check_frozen(range);
     /* Ranges are immutable, so that they should be initialized only once. */
     if (RANGE_EXCL(range) != Qnil) {
 	rb_name_err_raise("`initialize' called twice", range, ID2SYM(idInitialize));
@@ -91,7 +72,7 @@ range_modify(VALUE range)
  *     Range.new(begin, end, exclude_end=false)    -> rng
  *
  *  Constructs a range using the given +begin+ and +end+. If the +exclude_end+
- *  parameter is omitted or is <code>false</code>, the +rng+ will include
+ *  parameter is omitted or is <code>false</code>, the range will include
  *  the end object; otherwise, it will be excluded.
  */
 
@@ -249,11 +230,11 @@ range_hash(VALUE range)
     hash = rb_hash_uint(hash, EXCL(range) << 24);
     hash = rb_hash_end(hash);
 
-    return LONG2FIX(hash);
+    return ST2FIX(hash);
 }
 
 static void
-range_each_func(VALUE range, rb_block_call_func *func, VALUE arg)
+range_each_func(VALUE range, int (*func)(VALUE, VALUE), VALUE arg)
 {
     int c;
     VALUE b = RANGE_BEG(range);
@@ -262,21 +243,21 @@ range_each_func(VALUE range, rb_block_call_func *func, VALUE arg)
 
     if (EXCL(range)) {
 	while (r_less(v, e) < 0) {
-	    (*func) (v, arg, 0, 0, 0);
+	    if ((*func)(v, arg)) break;
 	    v = rb_funcallv(v, id_succ, 0, 0);
 	}
     }
     else {
 	while ((c = r_less(v, e)) <= 0) {
-	    (*func) (v, arg, 0, 0, 0);
+	    if ((*func)(v, arg)) break;
 	    if (!c) break;
 	    v = rb_funcallv(v, id_succ, 0, 0);
 	}
     }
 }
 
-static VALUE
-sym_step_i(RB_BLOCK_CALL_FUNC_ARGLIST(i, arg))
+static int
+sym_step_i(VALUE i, VALUE arg)
 {
     VALUE *iter = (VALUE *)arg;
 
@@ -290,11 +271,11 @@ sym_step_i(RB_BLOCK_CALL_FUNC_ARGLIST(i, arg))
 	rb_yield(rb_str_intern(i));
 	iter[0] = iter[1];
     }
-    return Qnil;
+    return 0;
 }
 
-static VALUE
-step_i(RB_BLOCK_CALL_FUNC_ARGLIST(i, arg))
+static int
+step_i(VALUE i, VALUE arg)
 {
     VALUE *iter = (VALUE *)arg;
 
@@ -308,7 +289,7 @@ step_i(RB_BLOCK_CALL_FUNC_ARGLIST(i, arg))
 	rb_yield(i);
 	iter[0] = iter[1];
     }
-    return Qnil;
+    return 0;
 }
 
 static int
@@ -367,16 +348,22 @@ range_step_size(VALUE range, VALUE args, VALUE eobj)
 }
 
 /*
+ *  Document-method: Range#step
+ *  Document-method: Range#%
  *  call-seq:
  *     rng.step(n=1) {| obj | block }    -> rng
  *     rng.step(n=1)                     -> an_enumerator
+ *     rng.step(n=1)                     -> an_arithmetic_sequence
+ *     rng % n                           -> an_enumerator
+ *     rng % n                           -> an_arithmetic_sequence
  *
  *  Iterates over the range, passing each <code>n</code>th element to the block.
  *  If begin and end are numeric, +n+ is added for each iteration.
- *  Otherwise <code>step</code> invokes <code>succ</code> to iterate through
- *  range elements.
+ *  Otherwise #step invokes #succ to iterate through range elements.
  *
  *  If no block is given, an enumerator is returned instead.
+ *  Especially, the enumerator is an Enumerator::ArithmeticSequence
+ *  if begin and end of the range are numeric.
  *
  *    range = Xs.new(1)..Xs.new(10)
  *    range.step(2) {|x| puts x}
@@ -405,19 +392,35 @@ range_step(int argc, VALUE *argv, VALUE range)
 {
     VALUE b, e, step, tmp;
 
-    RETURN_SIZED_ENUMERATOR(range, argc, argv, range_step_size);
-
     b = RANGE_BEG(range);
     e = RANGE_END(range);
-    if (argc == 0) {
-	step = INT2FIX(1);
-    }
-    else {
-	rb_scan_args(argc, argv, "01", &step);
-	step = check_step_domain(step);
+    step = (!rb_check_arity(argc, 0, 1) ? INT2FIX(1) : argv[0]);
+
+    if (!rb_block_given_p()) {
+        const VALUE b_num_p = rb_obj_is_kind_of(b, rb_cNumeric);
+        const VALUE e_num_p = rb_obj_is_kind_of(e, rb_cNumeric);
+        if ((b_num_p && (NIL_P(e) || e_num_p)) || (NIL_P(b) && e_num_p)) {
+            return rb_arith_seq_new(range, ID2SYM(rb_frame_this_func()), argc, argv,
+                    range_step_size, b, e, step, EXCL(range));
+        }
+
+        RETURN_SIZED_ENUMERATOR(range, argc, argv, range_step_size);
     }
 
-    if (FIXNUM_P(b) && FIXNUM_P(e) && FIXNUM_P(step)) { /* fixnums are special */
+    step = check_step_domain(step);
+
+    if (FIXNUM_P(b) && NIL_P(e) && FIXNUM_P(step)) {
+	long i = FIX2LONG(b), unit = FIX2LONG(step);
+	do {
+	    rb_yield(LONG2FIX(i));
+	    i += unit;          /* FIXABLE+FIXABLE never overflow */
+	} while (FIXABLE(i));
+	b = LONG2NUM(i);
+
+	for (;; b = rb_big_plus(b, step))
+	    rb_yield(b);
+    }
+    else if (FIXNUM_P(b) && FIXNUM_P(e) && FIXNUM_P(step)) { /* fixnums are special */
 	long end = FIX2LONG(e);
 	long i, unit = FIX2LONG(step);
 
@@ -431,16 +434,20 @@ range_step(int argc, VALUE *argv, VALUE range)
 	}
 
     }
-    else if (SYMBOL_P(b) && SYMBOL_P(e)) { /* symbols are special */
-	VALUE args[2], iter[2];
-
-	args[0] = rb_sym2str(e);
-	args[1] = EXCL(range) ? Qtrue : Qfalse;
+    else if (SYMBOL_P(b) && (NIL_P(e) || SYMBOL_P(e))) { /* symbols are special */
+	VALUE iter[2];
 	iter[0] = INT2FIX(1);
 	iter[1] = step;
-	rb_block_call(rb_sym2str(b), rb_intern("upto"), 2, args, sym_step_i, (VALUE)iter);
+
+	b = rb_sym2str(b);
+	if (NIL_P(e)) {
+	    rb_str_upto_endless_each(b, sym_step_i, (VALUE)iter);
+	}
+	else {
+	    rb_str_upto_each(b, rb_sym2str(e), EXCL(range), sym_step_i, (VALUE)iter);
+	}
     }
-    else if (ruby_float_step(b, e, step, EXCL(range))) {
+    else if (ruby_float_step(b, e, step, EXCL(range), TRUE)) {
 	/* done */
     }
     else if (rb_obj_is_kind_of(b, rb_cNumeric) ||
@@ -450,7 +457,7 @@ range_step(int argc, VALUE *argv, VALUE range)
 	VALUE v = b;
 	int i = 0;
 
-	while (RTEST(rb_funcall(v, op, 1, e))) {
+	while (NIL_P(e) || RTEST(rb_funcall(v, op, 1, e))) {
 	    rb_yield(v);
 	    i++;
 	    v = rb_funcall(b, '+', 1, rb_funcall(INT2NUM(i), '*', 1, step));
@@ -460,14 +467,18 @@ range_step(int argc, VALUE *argv, VALUE range)
 	tmp = rb_check_string_type(b);
 
 	if (!NIL_P(tmp)) {
-	    VALUE args[2], iter[2];
+	    VALUE iter[2];
 
 	    b = tmp;
-	    args[0] = e;
-	    args[1] = EXCL(range) ? Qtrue : Qfalse;
 	    iter[0] = INT2FIX(1);
 	    iter[1] = step;
-	    rb_block_call(b, rb_intern("upto"), 2, args, step_i, (VALUE)iter);
+
+	    if (NIL_P(e)) {
+		rb_str_upto_endless_each(b, step_i, (VALUE)iter);
+	    }
+	    else {
+		rb_str_upto_each(b, e, EXCL(range), step_i, (VALUE)iter);
+	    }
 	}
 	else {
 	    VALUE args[2];
@@ -484,6 +495,12 @@ range_step(int argc, VALUE *argv, VALUE range)
     return range;
 }
 
+static VALUE
+range_percent_step(VALUE range, VALUE step)
+{
+    return range_step(1, &step, range);
+}
+
 #if SIZEOF_DOUBLE == 8 && defined(HAVE_INT64_T)
 union int64_double {
     int64_t i;
@@ -516,10 +533,72 @@ double_as_int64(double d)
 static int
 is_integer_p(VALUE v)
 {
-    VALUE is_int = rb_check_funcall(v, id_integer_p, 0, 0);
+    ID id_integer_p;
+    VALUE is_int;
+    CONST_ID(id_integer_p, "integer?");
+    is_int = rb_check_funcall(v, id_integer_p, 0, 0);
     return RTEST(is_int) && is_int != Qundef;
 }
 
+static VALUE
+bsearch_integer_range(VALUE beg, VALUE end, int excl)
+{
+    VALUE satisfied = Qnil;
+    int smaller;
+
+#define BSEARCH_CHECK(expr) \
+    do { \
+	VALUE val = (expr); \
+	VALUE v = rb_yield(val); \
+	if (FIXNUM_P(v)) { \
+	    if (v == INT2FIX(0)) return val; \
+	    smaller = (SIGNED_VALUE)v < 0; \
+	} \
+	else if (v == Qtrue) { \
+	    satisfied = val; \
+	    smaller = 1; \
+	} \
+	else if (v == Qfalse || v == Qnil) { \
+	    smaller = 0; \
+	} \
+	else if (rb_obj_is_kind_of(v, rb_cNumeric)) { \
+	    int cmp = rb_cmpint(rb_funcall(v, id_cmp, 1, INT2FIX(0)), v, INT2FIX(0)); \
+	    if (!cmp) return val; \
+	    smaller = cmp < 0; \
+	} \
+	else { \
+	    rb_raise(rb_eTypeError, "wrong argument type %"PRIsVALUE \
+		     " (must be numeric, true, false or nil)", \
+		     rb_obj_class(v)); \
+	} \
+    } while (0)
+
+    VALUE low = rb_to_int(beg);
+    VALUE high = rb_to_int(end);
+    VALUE mid, org_high;
+    ID id_div;
+    CONST_ID(id_div, "div");
+
+    if (excl) high = rb_funcall(high, '-', 1, INT2FIX(1));
+    org_high = high;
+
+    while (rb_cmpint(rb_funcall(low, id_cmp, 1, high), low, high) < 0) {
+	mid = rb_funcall(rb_funcall(high, '+', 1, low), id_div, 1, INT2FIX(2));
+	BSEARCH_CHECK(mid);
+	if (smaller) {
+	    high = mid;
+	}
+	else {
+	    low = rb_funcall(mid, '+', 1, INT2FIX(1));
+	}
+    }
+    if (rb_equal(low, org_high)) {
+	BSEARCH_CHECK(low);
+	if (!smaller) return Qnil;
+    }
+    return satisfied;
+}
+
 /*
  *  call-seq:
  *     rng.bsearch {|obj| block }  -> value
@@ -592,33 +671,6 @@ range_bsearch(VALUE range)
      * (-1...0.0).bsearch to yield -0.0.
      */
 
-#define BSEARCH_CHECK(expr) \
-    do { \
-	VALUE val = (expr); \
-	VALUE v = rb_yield(val); \
-	if (FIXNUM_P(v)) { \
-	    if (v == INT2FIX(0)) return val; \
-	    smaller = (SIGNED_VALUE)v < 0; \
-	} \
-	else if (v == Qtrue) { \
-	    satisfied = val; \
-	    smaller = 1; \
-	} \
-	else if (v == Qfalse || v == Qnil) { \
-	    smaller = 0; \
-	} \
-	else if (rb_obj_is_kind_of(v, rb_cNumeric)) { \
-	    int cmp = rb_cmpint(rb_funcall(v, id_cmp, 1, INT2FIX(0)), v, INT2FIX(0)); \
-	    if (!cmp) return val; \
-	    smaller = cmp < 0; \
-	} \
-	else { \
-	    rb_raise(rb_eTypeError, "wrong argument type %"PRIsVALUE \
-		     " (must be numeric, true, false or nil)", \
-		     rb_obj_class(v)); \
-	} \
-    } while (0)
-
 #define BSEARCH(conv) \
     do { \
 	RETURN_ENUMERATOR(range, 0, 0); \
@@ -654,35 +706,39 @@ range_bsearch(VALUE range)
     }
 #if SIZEOF_DOUBLE == 8 && defined(HAVE_INT64_T)
     else if (RB_TYPE_P(beg, T_FLOAT) || RB_TYPE_P(end, T_FLOAT)) {
-	int64_t low  = double_as_int64(RFLOAT_VALUE(rb_Float(beg)));
-	int64_t high = double_as_int64(RFLOAT_VALUE(rb_Float(end)));
+	int64_t low  = double_as_int64(NIL_P(beg) ? -HUGE_VAL : RFLOAT_VALUE(rb_Float(beg)));
+	int64_t high = double_as_int64(NIL_P(end) ?  HUGE_VAL : RFLOAT_VALUE(rb_Float(end)));
 	int64_t mid, org_high;
 	BSEARCH(int64_as_double_to_num);
     }
 #endif
     else if (is_integer_p(beg) && is_integer_p(end)) {
-	VALUE low = rb_to_int(beg);
-	VALUE high = rb_to_int(end);
-	VALUE mid, org_high;
 	RETURN_ENUMERATOR(range, 0, 0);
-	if (EXCL(range)) high = rb_funcall(high, '-', 1, INT2FIX(1));
-	org_high = high;
-
-	while (rb_cmpint(rb_funcall(low, id_cmp, 1, high), low, high) < 0) {
-	    mid = rb_funcall(rb_funcall(high, '+', 1, low), id_div, 1, INT2FIX(2));
+	return bsearch_integer_range(beg, end, EXCL(range));
+    }
+    else if (is_integer_p(beg) && NIL_P(end)) {
+	VALUE diff = LONG2FIX(1);
+	RETURN_ENUMERATOR(range, 0, 0);
+	while (1) {
+	    VALUE mid = rb_funcall(beg, '+', 1, diff);
 	    BSEARCH_CHECK(mid);
 	    if (smaller) {
-		high = mid;
+		return bsearch_integer_range(beg, mid, 0);
 	    }
-	    else {
-		low = rb_funcall(mid, '+', 1, INT2FIX(1));
+	    diff = rb_funcall(diff, '*', 1, LONG2FIX(2));
+	}
+    }
+    else if (NIL_P(beg) && is_integer_p(end)) {
+	VALUE diff = LONG2FIX(-1);
+	RETURN_ENUMERATOR(range, 0, 0);
+	while (1) {
+	    VALUE mid = rb_funcall(end, '+', 1, diff);
+	    BSEARCH_CHECK(mid);
+	    if (!smaller) {
+		return bsearch_integer_range(mid, end, 0);
 	    }
+	    diff = rb_funcall(diff, '*', 1, LONG2FIX(2));
 	}
-	if (rb_equal(low, org_high)) {
-	    BSEARCH_CHECK(low);
-	    if (!smaller) return Qnil;
-	}
-	return satisfied;
     }
     else {
 	rb_raise(rb_eTypeError, "can't do binary search for %s", rb_obj_classname(beg));
@@ -690,18 +746,18 @@ range_bsearch(VALUE range)
     return range;
 }
 
-static VALUE
-each_i(RB_BLOCK_CALL_FUNC_ARGLIST(v, arg))
+static int
+each_i(VALUE v, VALUE arg)
 {
     rb_yield(v);
-    return Qnil;
+    return 0;
 }
 
-static VALUE
-sym_each_i(RB_BLOCK_CALL_FUNC_ARGLIST(v, arg))
+static int
+sym_each_i(VALUE v, VALUE arg)
 {
     rb_yield(rb_str_intern(v));
-    return Qnil;
+    return 0;
 }
 
 /*
@@ -720,12 +776,41 @@ static VALUE
 range_size(VALUE range)
 {
     VALUE b = RANGE_BEG(range), e = RANGE_END(range);
-    if (rb_obj_is_kind_of(b, rb_cNumeric) && rb_obj_is_kind_of(e, rb_cNumeric)) {
-	return ruby_num_interval_step_size(b, e, INT2FIX(1), EXCL(range));
+    if (rb_obj_is_kind_of(b, rb_cNumeric)) {
+        if (rb_obj_is_kind_of(e, rb_cNumeric)) {
+	    return ruby_num_interval_step_size(b, e, INT2FIX(1), EXCL(range));
+        }
+        if (NIL_P(e)) {
+            return DBL2NUM(HUGE_VAL);
+        }
     }
+    else if (NIL_P(b)) {
+        return DBL2NUM(HUGE_VAL);
+    }
+
     return Qnil;
 }
 
+/*
+ *  call-seq:
+ *     rng.to_a                   -> array
+ *     rng.entries                -> array
+ *
+ *  Returns an array containing the items in the range.
+ *
+ *    (1..7).to_a  #=> [1, 2, 3, 4, 5, 6, 7]
+ *    (1..).to_a   #=> RangeError: cannot convert endless range to an array
+ */
+
+static VALUE
+range_to_a(VALUE range)
+{
+    if (NIL_P(RANGE_END(range))) {
+	rb_raise(rb_eRangeError, "cannot convert endless range to an array");
+    }
+    return rb_call_super(0, 0);
+}
+
 static VALUE
 range_enum_size(VALUE range, VALUE args, VALUE eobj)
 {
@@ -757,45 +842,105 @@ static VALUE
 range_each(VALUE range)
 {
     VALUE beg, end;
+    long i, lim;
 
     RETURN_SIZED_ENUMERATOR(range, 0, 0, range_enum_size);
 
     beg = RANGE_BEG(range);
     end = RANGE_END(range);
 
-    if (FIXNUM_P(beg) && FIXNUM_P(end)) { /* fixnums are special */
-	long lim = FIX2LONG(end);
-	long i;
-
+    if (FIXNUM_P(beg) && NIL_P(end)) {
+      fixnum_endless:
+	i = FIX2LONG(beg);
+	while (FIXABLE(i)) {
+	    rb_yield(LONG2FIX(i++));
+	}
+	beg = LONG2NUM(i);
+      bignum_endless:
+	for (;; beg = rb_big_plus(beg, INT2FIX(1)))
+	    rb_yield(beg);
+    }
+    else if (FIXNUM_P(beg) && FIXNUM_P(end)) { /* fixnums are special */
+      fixnum_loop:
+	lim = FIX2LONG(end);
 	if (!EXCL(range))
 	    lim += 1;
 	for (i = FIX2LONG(beg); i < lim; i++) {
 	    rb_yield(LONG2FIX(i));
 	}
     }
-    else if (SYMBOL_P(beg) && SYMBOL_P(end)) { /* symbols are special */
-	VALUE args[2];
-
-	args[0] = rb_sym2str(end);
-	args[1] = EXCL(range) ? Qtrue : Qfalse;
-	rb_block_call(rb_sym2str(beg), rb_intern("upto"), 2, args, sym_each_i, 0);
+    else if (RB_INTEGER_TYPE_P(beg) && (NIL_P(end) || RB_INTEGER_TYPE_P(end))) {
+	if (SPECIAL_CONST_P(end) || RBIGNUM_POSITIVE_P(end)) { /* end >= FIXNUM_MIN */
+	    if (!FIXNUM_P(beg)) {
+		if (RBIGNUM_NEGATIVE_P(beg)) {
+		    do {
+			rb_yield(beg);
+		    } while (!FIXNUM_P(beg = rb_big_plus(beg, INT2FIX(1))));
+		    if (NIL_P(end)) goto fixnum_endless;
+		    if (FIXNUM_P(end)) goto fixnum_loop;
+		}
+		else {
+		    if (NIL_P(end)) goto bignum_endless;
+		    if (FIXNUM_P(end)) return range;
+		}
+	    }
+	    if (FIXNUM_P(beg)) {
+		i = FIX2LONG(beg);
+		do {
+		    rb_yield(LONG2FIX(i));
+		} while (POSFIXABLE(++i));
+		beg = LONG2NUM(i);
+	    }
+	    ASSUME(!FIXNUM_P(beg));
+	    ASSUME(!SPECIAL_CONST_P(end));
+	}
+	if (!FIXNUM_P(beg) && RBIGNUM_SIGN(beg) == RBIGNUM_SIGN(end)) {
+	    if (EXCL(range)) {
+		while (rb_big_cmp(beg, end) == INT2FIX(-1)) {
+		    rb_yield(beg);
+		    beg = rb_big_plus(beg, INT2FIX(1));
+		}
+	    }
+	    else {
+		VALUE c;
+		while ((c = rb_big_cmp(beg, end)) != INT2FIX(1)) {
+		    rb_yield(beg);
+		    if (c == INT2FIX(0)) break;
+		    beg = rb_big_plus(beg, INT2FIX(1));
+		}
+	    }
+	}
+    }
+    else if (SYMBOL_P(beg) && (NIL_P(end) || SYMBOL_P(end))) { /* symbols are special */
+	beg = rb_sym2str(beg);
+	if (NIL_P(end)) {
+	    rb_str_upto_endless_each(beg, sym_each_i, 0);
+	}
+	else {
+	    rb_str_upto_each(beg, rb_sym2str(end), EXCL(range), sym_each_i, 0);
+	}
     }
     else {
 	VALUE tmp = rb_check_string_type(beg);
 
 	if (!NIL_P(tmp)) {
-	    VALUE args[2];
-
-	    args[0] = end;
-	    args[1] = EXCL(range) ? Qtrue : Qfalse;
-	    rb_block_call(tmp, rb_intern("upto"), 2, args, each_i, 0);
+	    if (!NIL_P(end)) {
+		rb_str_upto_each(tmp, end, EXCL(range), each_i, 0);
+	    }
+	    else {
+		rb_str_upto_endless_each(tmp, each_i, 0);
+	    }
 	}
 	else {
 	    if (!discrete_object_p(beg)) {
 		rb_raise(rb_eTypeError, "can't iterate from %s",
 			 rb_obj_classname(beg));
 	    }
-	    range_each_func(range, each_i, 0);
+	    if (!NIL_P(end))
+		range_each_func(range, each_i, 0);
+	    else
+		for (;; beg = rb_funcallv(beg, id_succ, 0, 0))
+		    rb_yield(beg);
 	}
     }
     return range;
@@ -867,6 +1012,9 @@ range_first(int argc, VALUE *argv, VALUE range)
 {
     VALUE n, ary[2];
 
+    if (NIL_P(RANGE_BEG(range))) {
+        rb_raise(rb_eRangeError, "cannot get the first element of beginless range");
+    }
     if (argc == 0) return RANGE_BEG(range);
 
     rb_scan_args(argc, argv, "1", &n);
@@ -877,6 +1025,58 @@ range_first(int argc, VALUE *argv, VALUE range)
     return ary[1];
 }
 
+static VALUE
+rb_int_range_last(int argc, VALUE *argv, VALUE range)
+{
+    static const VALUE ONE = INT2FIX(1);
+
+    VALUE b, e, len_1, len, nv, ary;
+    int x;
+    long n;
+
+    assert(argc > 0);
+
+    b = RANGE_BEG(range);
+    e = RANGE_END(range);
+    assert(RB_INTEGER_TYPE_P(b) && RB_INTEGER_TYPE_P(e));
+
+    x = EXCL(range);
+
+    len_1 = rb_int_minus(e, b);
+    if (FIXNUM_ZERO_P(len_1) || rb_num_negative_p(len_1)) {
+        return rb_ary_new_capa(0);
+    }
+
+    if (x) {
+        e = rb_int_minus(e, ONE);
+        len = len_1;
+    }
+    else {
+        len = rb_int_plus(len_1, ONE);
+    }
+
+    rb_scan_args(argc, argv, "1", &nv);
+    n = NUM2LONG(nv);
+    if (n < 0) {
+        rb_raise(rb_eArgError, "negative array size");
+    }
+
+    nv = LONG2NUM(n);
+    if (RTEST(rb_int_gt(nv, len))) {
+        nv = len;
+        n = NUM2LONG(nv);
+    }
+
+    ary = rb_ary_new_capa(n);
+    b = rb_int_minus(e, nv);
+    while (n) {
+        b = rb_int_plus(b, ONE);
+        rb_ary_push(ary, b);
+        --n;
+    }
+
+    return ary;
+}
 
 /*
  *  call-seq:
@@ -898,7 +1098,19 @@ range_first(int argc, VALUE *argv, VALUE range)
 static VALUE
 range_last(int argc, VALUE *argv, VALUE range)
 {
+    VALUE b, e;
+
+    if (NIL_P(RANGE_END(range))) {
+        rb_raise(rb_eRangeError, "cannot get the last element of endless range");
+    }
     if (argc == 0) return RANGE_END(range);
+
+    b = RANGE_BEG(range);
+    e = RANGE_END(range);
+    if (RB_INTEGER_TYPE_P(b) && RB_INTEGER_TYPE_P(e) &&
+        RB_LIKELY(rb_method_basic_definition_p(rb_cRange, idEach))) {
+        return rb_int_range_last(argc, argv, range);
+    }
     return rb_ary_last(argc, argv, rb_Array(range));
 }
 
@@ -925,15 +1137,19 @@ static VALUE
 range_min(int argc, VALUE *argv, VALUE range)
 {
     if (rb_block_given_p()) {
+        if (NIL_P(RANGE_END(range))) {
+            rb_raise(rb_eRangeError, "cannot get the minimum of endless range with custom comparison method");
+        }
 	return rb_call_super(argc, argv);
     }
     else if (argc != 0) {
 	return range_first(argc, argv, range);
     }
     else {
+	struct cmp_opt_data cmp_opt = { 0, 0 };
 	VALUE b = RANGE_BEG(range);
 	VALUE e = RANGE_END(range);
-	int c = rb_cmpint(rb_funcall(b, id_cmp, 1, e), b, e);
+	int c = NIL_P(e) ? -1 : OPTIMIZED_CMP(b, e, cmp_opt);
 
 	if (c > 0 || (c == 0 && EXCL(range)))
 	    return Qnil;
@@ -964,32 +1180,58 @@ range_max(int argc, VALUE *argv, VALUE range)
     VALUE e = RANGE_END(range);
     int nm = FIXNUM_P(e) || rb_obj_is_kind_of(e, rb_cNumeric);
 
+    if (NIL_P(RANGE_END(range))) {
+	rb_raise(rb_eRangeError, "cannot get the maximum of endless range");
+    }
+
     if (rb_block_given_p() || (EXCL(range) && !nm) || argc) {
-	return rb_call_super(argc, argv);
+        return rb_call_super(argc, argv);
     }
     else {
-	VALUE b = RANGE_BEG(range);
-	int c = rb_cmpint(rb_funcall(b, id_cmp, 1, e), b, e);
+        struct cmp_opt_data cmp_opt = { 0, 0 };
+        VALUE b = RANGE_BEG(range);
+        int c = OPTIMIZED_CMP(b, e, cmp_opt);
 
-	if (c > 0)
-	    return Qnil;
-	if (EXCL(range)) {
-	    if (!FIXNUM_P(e) && !rb_obj_is_kind_of(e, rb_cInteger)) {
-		rb_raise(rb_eTypeError, "cannot exclude non Integer end value");
-	    }
-	    if (c == 0) return Qnil;
-	    if (!FIXNUM_P(b) && !rb_obj_is_kind_of(b,rb_cInteger)) {
-		rb_raise(rb_eTypeError, "cannot exclude end value with non Integer begin value");
-	    }
-	    if (FIXNUM_P(e)) {
-		return LONG2NUM(FIX2LONG(e) - 1);
-	    }
-	    return rb_funcall(e, '-', 1, INT2FIX(1));
-	}
-	return e;
+        if (c > 0)
+            return Qnil;
+        if (EXCL(range)) {
+            if (!RB_INTEGER_TYPE_P(e)) {
+                rb_raise(rb_eTypeError, "cannot exclude non Integer end value");
+            }
+            if (c == 0) return Qnil;
+            if (!RB_INTEGER_TYPE_P(b)) {
+                rb_raise(rb_eTypeError, "cannot exclude end value with non Integer begin value");
+            }
+            if (FIXNUM_P(e)) {
+                return LONG2NUM(FIX2LONG(e) - 1);
+            }
+            return rb_funcall(e, '-', 1, INT2FIX(1));
+        }
+        return e;
     }
 }
 
+/*
+ *  call-seq:
+ *     rng.minmax                       -> [obj, obj]
+ *     rng.minmax {| a,b | block }      -> [obj, obj]
+ *
+ *  Returns a two element array which contains the minimum and the
+ *  maximum value in the range.
+ *
+ *  Can be given an optional block to override the default comparison
+ *  method <code>a <=> b</code>.
+ */
+
+static VALUE
+range_minmax(VALUE range)
+{
+    if (rb_block_given_p()) {
+        return rb_call_super(0, NULL);
+    }
+    return rb_assoc_new(range_min(0, NULL, range), range_max(0, NULL, range));
+}
+
 int
 rb_range_values(VALUE range, VALUE *begp, VALUE *endp, int *exclp)
 {
@@ -1001,12 +1243,18 @@ rb_range_values(VALUE range, VALUE *begp, VALUE *endp, int *exclp)
 	e = RANGE_END(range);
 	excl = EXCL(range);
     }
+    else if (RTEST(rb_obj_is_kind_of(range, rb_cArithSeq))) {
+        return (int)Qfalse;
+    }
     else {
-	if (!rb_respond_to(range, id_beg)) return (int)Qfalse;
-	if (!rb_respond_to(range, id_end)) return (int)Qfalse;
-	b = rb_funcall(range, id_beg, 0);
-	e = rb_funcall(range, id_end, 0);
-	excl = RTEST(rb_funcall(range, rb_intern("exclude_end?"), 0));
+	VALUE x;
+	b = rb_check_funcall(range, id_beg, 0, 0);
+	if (b == Qundef) return (int)Qfalse;
+	e = rb_check_funcall(range, id_end, 0, 0);
+	if (e == Qundef) return (int)Qfalse;
+	x = rb_check_funcall(range, rb_intern("exclude_end?"), 0, 0);
+	if (x == Qundef) return (int)Qfalse;
+	excl = RTEST(x);
     }
     *begp = b;
     *endp = e;
@@ -1023,8 +1271,9 @@ rb_range_beg_len(VALUE range, long *begp, long *lenp, long len, int err)
 
     if (!rb_range_values(range, &b, &e, &excl))
 	return Qfalse;
-    beg = NUM2LONG(b);
-    end = NUM2LONG(e);
+    beg = NIL_P(b) ? 0 : NUM2LONG(b);
+    end = NIL_P(e) ? -1 : NUM2LONG(e);
+    if (NIL_P(e)) excl = 0;
     origbeg = beg;
     origend = end;
     if (beg < 0) {
@@ -1076,7 +1325,6 @@ range_to_s(VALUE range)
     str = rb_str_dup(str);
     rb_str_cat(str, "...", EXCL(range) ? 3 : 2);
     rb_str_append(str, str2);
-    OBJ_INFECT(str, range);
 
     return str;
 }
@@ -1084,17 +1332,22 @@ range_to_s(VALUE range)
 static VALUE
 inspect_range(VALUE range, VALUE dummy, int recur)
 {
-    VALUE str, str2;
+    VALUE str, str2 = Qundef;
 
     if (recur) {
 	return rb_str_new2(EXCL(range) ? "(... ... ...)" : "(... .. ...)");
     }
-    str = rb_inspect(RANGE_BEG(range));
-    str2 = rb_inspect(RANGE_END(range));
-    str = rb_str_dup(str);
+    if (!NIL_P(RANGE_BEG(range)) || NIL_P(RANGE_END(range))) {
+        str = rb_str_dup(rb_inspect(RANGE_BEG(range)));
+    }
+    else {
+        str = rb_str_new(0, 0);
+    }
     rb_str_cat(str, "...", EXCL(range) ? 3 : 2);
-    rb_str_append(str, str2);
-    OBJ_INFECT(str, range);
+    if (NIL_P(RANGE_BEG(range)) || !NIL_P(RANGE_END(range))) {
+        str2 = rb_inspect(RANGE_END(range));
+    }
+    if (str2 != Qundef) rb_str_append(str, str2);
 
     return str;
 }
@@ -1103,9 +1356,8 @@ inspect_range(VALUE range, VALUE dummy, int recur)
  * call-seq:
  *   rng.inspect  -> string
  *
- * Convert this range object to a printable form (using
- * <code>inspect</code> to convert the begin and end
- * objects).
+ * Convert this range object to a printable form (using #inspect to
+ * convert the begin and end objects).
  */
 
 
@@ -1115,6 +1367,8 @@ range_inspect(VALUE range)
     return rb_exec_recursive(inspect_range, range, 0);
 }
 
+static VALUE range_include_internal(VALUE range, VALUE val, int string_use_cover);
+
 /*
  *  call-seq:
  *     rng === obj       ->  true or false
@@ -1137,7 +1391,9 @@ range_inspect(VALUE range)
 static VALUE
 range_eqq(VALUE range, VALUE val)
 {
-    return rb_funcall(range, rb_intern("include?"), 1, val);
+    VALUE ret = range_include_internal(range, val, 1);
+    if (ret != Qundef) return ret;
+    return r_cover_p(range, RANGE_BEG(range), RANGE_END(range), val);
 }
 
 
@@ -1157,6 +1413,14 @@ range_eqq(VALUE range, VALUE val)
 
 static VALUE
 range_include(VALUE range, VALUE val)
+{
+    VALUE ret = range_include_internal(range, val, 0);
+    if (ret != Qundef) return ret;
+    return rb_call_super(1, &val);
+}
+
+static VALUE
+range_include_internal(VALUE range, VALUE val, int string_use_cover)
 {
     VALUE beg = RANGE_BEG(range);
     VALUE end = RANGE_END(range);
@@ -1168,18 +1432,38 @@ range_include(VALUE range, VALUE val)
 	!NIL_P(rb_check_to_integer(end, "to_int"))) {
 	return r_cover_p(range, beg, end, val);
     }
-    else if (RB_TYPE_P(beg, T_STRING) && RB_TYPE_P(end, T_STRING)) {
-	VALUE rb_str_include_range_p(VALUE beg, VALUE end, VALUE val, VALUE exclusive);
-	return rb_str_include_range_p(beg, end, val, RANGE_EXCL(range));
+    else if (RB_TYPE_P(beg, T_STRING) || RB_TYPE_P(end, T_STRING)) {
+        if (RB_TYPE_P(beg, T_STRING) && RB_TYPE_P(end, T_STRING)) {
+            if (string_use_cover) {
+                return r_cover_p(range, beg, end, val);
+            }
+            else {
+                VALUE rb_str_include_range_p(VALUE beg, VALUE end, VALUE val, VALUE exclusive);
+                return rb_str_include_range_p(beg, end, val, RANGE_EXCL(range));
+            }
+        }
+        else if (NIL_P(beg)) {
+	    VALUE r = rb_funcall(val, id_cmp, 1, end);
+	    if (NIL_P(r)) return Qfalse;
+	    if (rb_cmpint(r, val, end) <= 0) return Qtrue;
+	    return Qfalse;
+        }
+	else if (NIL_P(end)) {
+	    VALUE r = rb_funcall(beg, id_cmp, 1, val);
+	    if (NIL_P(r)) return Qfalse;
+	    if (rb_cmpint(r, beg, val) <= 0) return Qtrue;
+	    return Qfalse;
+	}
     }
-    /* TODO: ruby_frame->this_func = rb_intern("include?"); */
-    return rb_call_super(1, &val);
+    return Qundef;
 }
 
+static int r_cover_range_p(VALUE range, VALUE beg, VALUE end, VALUE val);
 
 /*
  *  call-seq:
- *     rng.cover?(obj)  ->  true or false
+ *     rng.cover?(obj)   ->  true or false
+ *     rng.cover?(range) ->  true or false
  *
  *  Returns <code>true</code> if +obj+ is between the begin and end of
  *  the range.
@@ -1187,9 +1471,25 @@ range_include(VALUE range, VALUE val)
  *  This tests <code>begin <= obj <= end</code> when #exclude_end? is +false+
  *  and <code>begin <= obj < end</code> when #exclude_end? is +true+.
  *
- *     ("a".."z").cover?("c")    #=> true
- *     ("a".."z").cover?("5")    #=> false
- *     ("a".."z").cover?("cc")   #=> true
+ *  If called with a Range argument, returns <code>true</code> when the
+ *  given range is covered by the receiver,
+ *  by comparing the begin and end values. If the argument can be treated as
+ *  a sequence, this method treats it that way. In the specific case of
+ *  <code>(a..b).cover?(c...d)</code> with <code>a <= c && b < d</code>,
+ *  the end of the sequence must be calculated, which may exhibit poor
+ *  performance if <code>c</code> is non-numeric.
+ *  Returns <code>false</code> if the begin value of the
+ *  range is larger than the end value. Also returns +false+ if one of the
+ *  internal calls to <code><=></code> returns +nil+ (indicating the objects
+ *  are not comparable).
+ *
+ *     ("a".."z").cover?("c")  #=> true
+ *     ("a".."z").cover?("5")  #=> false
+ *     ("a".."z").cover?("cc") #=> true
+ *     ("a".."z").cover?(1)    #=> false
+ *     (1..5).cover?(2..3)     #=> true
+ *     (1..5).cover?(0..6)     #=> false
+ *     (1..5).cover?(1...6)    #=> true
  */
 
 static VALUE
@@ -1199,15 +1499,57 @@ range_cover(VALUE range, VALUE val)
 
     beg = RANGE_BEG(range);
     end = RANGE_END(range);
+
+    if (rb_obj_is_kind_of(val, rb_cRange)) {
+        return RBOOL(r_cover_range_p(range, beg, end, val));
+    }
     return r_cover_p(range, beg, end, val);
 }
 
+static VALUE
+r_call_max(VALUE r)
+{
+    return rb_funcallv(r, rb_intern("max"), 0, 0);
+}
+
+static int
+r_cover_range_p(VALUE range, VALUE beg, VALUE end, VALUE val)
+{
+    VALUE val_beg, val_end, val_max;
+    int cmp_end;
+
+    val_beg = RANGE_BEG(val);
+    val_end = RANGE_END(val);
+
+    if (!NIL_P(end) && NIL_P(val_end)) return FALSE;
+    if (!NIL_P(beg) && NIL_P(val_beg)) return FALSE;
+    if (!NIL_P(val_beg) && !NIL_P(val_end) && r_less(val_beg, val_end) > -EXCL(val)) return FALSE;
+    if (!NIL_P(val_beg) && !r_cover_p(range, beg, end, val_beg)) return FALSE;
+
+    cmp_end = r_less(end, val_end);
+
+    if (EXCL(range) == EXCL(val)) {
+        return cmp_end >= 0;
+    }
+    else if (EXCL(range)) {
+        return cmp_end > 0;
+    }
+    else if (cmp_end >= 0) {
+        return TRUE;
+    }
+
+    val_max = rb_rescue2(r_call_max, val, NULL, Qnil, rb_eTypeError, (VALUE)0);
+    if (val_max == Qnil) return FALSE;
+
+    return r_less(end, val_max) >= 0;
+}
+
 static VALUE
 r_cover_p(VALUE range, VALUE beg, VALUE end, VALUE val)
 {
-    if (r_less(beg, val) <= 0) {
+    if (NIL_P(beg) || r_less(beg, val) <= 0) {
 	int excl = EXCL(range);
-	if (r_less(val, end) <= -excl)
+	if (NIL_P(end) || r_less(val, end) <= -excl)
 	    return Qtrue;
     }
     return Qfalse;
@@ -1230,26 +1572,31 @@ range_dumper(VALUE range)
 static VALUE
 range_loader(VALUE range, VALUE obj)
 {
+    VALUE beg, end, excl;
+
     if (!RB_TYPE_P(obj, T_OBJECT) || RBASIC(obj)->klass != rb_cObject) {
         rb_raise(rb_eTypeError, "not a dumped range object");
     }
 
     range_modify(range);
-    RANGE_SET_BEG(range, rb_ivar_get(obj, id_beg));
-    RANGE_SET_END(range, rb_ivar_get(obj, id_end));
-    RANGE_SET_EXCL(range, rb_ivar_get(obj, id_excl));
+    beg = rb_ivar_get(obj, id_beg);
+    end = rb_ivar_get(obj, id_end);
+    excl = rb_ivar_get(obj, id_excl);
+    if (!NIL_P(excl)) {
+	range_init(range, beg, end, RBOOL(RTEST(excl)));
+    }
     return range;
 }
 
 static VALUE
 range_alloc(VALUE klass)
 {
-  /* rb_struct_alloc_noinit itself should not be used because
-   * rb_marshal_define_compat uses equality of allocation function */
+    /* rb_struct_alloc_noinit itself should not be used because
+     * rb_marshal_define_compat uses equality of allocation function */
     return rb_struct_alloc_noinit(klass);
 }
 
-/*  A <code>Range</code> represents an interval---a set of values with a
+/*  A Range represents an interval---a set of values with a
  *  beginning and an end. Ranges may be constructed using the
  *  <em>s</em><code>..</code><em>e</em> and
  *  <em>s</em><code>...</code><em>e</em> literals, or with
@@ -1263,6 +1610,42 @@ range_alloc(VALUE klass)
  *     ('a'..'e').to_a    #=> ["a", "b", "c", "d", "e"]
  *     ('a'...'e').to_a   #=> ["a", "b", "c", "d"]
  *
+ *  == Beginless/Endless Ranges
+ *
+ *  A "beginless range" and "endless range" represents a semi-infinite
+ *  range.  Literal notation for a beginless range is:
+ *
+ *     (..1)
+ *     # or
+ *     (...1)
+ *
+ *  Literal notation for an endless range is:
+ *
+ *     (1..)
+ *     # or similarly
+ *     (1...)
+ *
+ *  Which is equivalent to
+ *
+ *     (1..nil)  # or similarly (1...nil)
+ *     Range.new(1, nil) # or Range.new(1, nil, true)
+ *
+ *  Beginless/endless ranges are useful, for example, for idiomatic
+ *  slicing of arrays:
+ *
+ *    [1, 2, 3, 4, 5][...2]   # => [1, 2]
+ *    [1, 2, 3, 4, 5][2...]   # => [3, 4, 5]
+ *
+ *  Some implementation details:
+ *
+ *  * +begin+ of beginless range and +end+ of endless range are +nil+;
+ *  * +each+ of beginless range raises an exception;
+ *  * +each+ of endless range enumerates infinite sequence (may be
+ *    useful in combination with Enumerable#take_while or similar
+ *    methods);
+ *  * <code>(1..)</code> and <code>(1...)</code> are not equal,
+ *    although technically representing the same sequence.
+ *
  *  == Custom Objects in Ranges
  *
  *  Ranges can be constructed using any objects that can be compared
@@ -1315,8 +1698,6 @@ Init_Range(void)
     id_beg = rb_intern("begin");
     id_end = rb_intern("end");
     id_excl = rb_intern("excl");
-    id_integer_p = rb_intern("integer?");
-    id_div = rb_intern("div");
 
     rb_cRange = rb_struct_define_without_accessor(
         "Range", rb_cObject, range_alloc,
@@ -1332,6 +1713,7 @@ Init_Range(void)
     rb_define_method(rb_cRange, "hash", range_hash, 0);
     rb_define_method(rb_cRange, "each", range_each, 0);
     rb_define_method(rb_cRange, "step", range_step, -1);
+    rb_define_method(rb_cRange, "%", range_percent_step, 1);
     rb_define_method(rb_cRange, "bsearch", range_bsearch, 0);
     rb_define_method(rb_cRange, "begin", range_begin, 0);
     rb_define_method(rb_cRange, "end", range_end, 0);
@@ -1339,7 +1721,10 @@ Init_Range(void)
     rb_define_method(rb_cRange, "last", range_last, -1);
     rb_define_method(rb_cRange, "min", range_min, -1);
     rb_define_method(rb_cRange, "max", range_max, -1);
+    rb_define_method(rb_cRange, "minmax", range_minmax, 0);
     rb_define_method(rb_cRange, "size", range_size, 0);
+    rb_define_method(rb_cRange, "to_a", range_to_a, 0);
+    rb_define_method(rb_cRange, "entries", range_to_a, 0);
     rb_define_method(rb_cRange, "to_s", range_to_s, 0);
     rb_define_method(rb_cRange, "inspect", range_inspect, 0);
 
